"""
Cron Tasks.

Background tasks that run a worker server.
These tasks can run even while the user is not online,
not making any active HTTP requests, or not using the application.
"""

import asyncio
import logging
from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Any, Callable, ClassVar, Optional, TypedDict
from unittest import mock

import celery
import celery.schedules
import httpx

from beanie.odm.queries.find import FindMany
from fastapi import status as rest_status

from sap.loggers import logger
from sap.settings import SapSettings


class FetchStrategy(IntEnum):
    """Define if new or old data should be fetched."""

    NEW: int = 1
    OLD: int = 2


class CronResponseStatus(Enum):
    """Status of the crontask after it finish running."""

    SUCCESS: str = "Success"
    ABORTED: str = "Aborted"
    ERROR: str = "Error"


class CronResponse(TypedDict, total=False):
    """Define a standard cron task response."""

    error: dict[str, str]
    result: dict[str, int]
    status: str


@dataclass
class CronStat:
    """Metric that gives insight into data to be processed by a cron."""

    name: str
    value: int


class BaseCronTask(celery.Task):  # type: ignore
    """Define how cron task classes should be structured."""

    expires = 60 * 60  # automatically expires if not run within 1 hour
    time_limit = 60 * 60 * 3  # default to 3 hours, automatically kill the task if exceed the limit
    name: str
    args: list[Any] = []
    kwargs: dict[str, Any] = {}
    schedule: celery.schedules.crontab
    logger: logging.Logger = logger

    def __init__(self, **kw_args: Any) -> None:
        """Initialize the cron task."""
        self.name = self.get_name()
        for k, v in kw_args.items():
            setattr(self, k, v)

    @classmethod
    def get_name(cls) -> str:
        """Get Name of the current Task."""
        return cls.__module__.split(".crons", maxsplit=1)[0] + "." + str(cls.__name__)

    def get_queryset(self, *, batch_size: Optional[int] = None, **kwargs: Any) -> Any:
        """Fetch the list of elements to process."""
        raise NotImplementedError

    async def get_stats(self) -> list[CronStat]:
        """Give stats about the number of elements left to process."""
        raise NotImplementedError

    async def process(self, *, batch_size: int = 100, **kwargs: Any) -> Any:
        """Run the cron task and process elements."""
        raise NotImplementedError

    async def handle_process(self, *args: Any, **kwargs: Any) -> CronResponse:
        """Run the task and save meta info to Airtable."""
        response: CronResponse

        try:
            result = await self.process(**self.kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            if not SapSettings.is_env_prod:
                raise
            self.logger.exception(exc)
            response = {
                "error": {"class": exc.__class__.__name__, "message": str(exc)},
                "status": CronResponseStatus.ERROR.value,
            }
        else:
            response = {"result": result, "status": CronResponseStatus.SUCCESS.value}

        return response

    async def test_process(self, filter_queryset: Callable[[FindMany[Any]], FindMany[Any]]) -> CronResponse:
        """Call this method to launch the task in test cases.

        filter_queryset: This allows you to run an extra filtering on the data being processing.
        Useful if you want to limit the data processing to a specific sample.
        """
        original_get_queryset = self.get_queryset

        def mock_get_queryset(batch_size: Optional[int] = None, **kwargs: Any) -> FindMany[Any]:
            """Replace the normal filter by a new test filter."""
            queryset = original_get_queryset(batch_size=batch_size, **kwargs)
            return filter_queryset(queryset)

        with mock.patch.object(self, "get_queryset", side_effect=mock_get_queryset):
            return await self.handle_process()

    def run(self, *args: Any, **kwargs: Any) -> Any:
        """Run the task."""
        logger.debug("Running task=%s args=%s kwargs=%s", self.get_name(), str(args), str(kwargs))
        return asyncio.run(self.handle_process(*args, **kwargs))


class CronStorage:
    """
    Interface that store cron results in a database.

    Results can be used to collect statistics about cron runs.
    """

    task: BaseCronTask
    task_id: Optional[str] = None
    task_name: str

    def __init__(self, task: BaseCronTask):
        """Initialize the storage."""
        self.task = task
        self.task_name = task.get_name()

    async def record_task(self) -> None:
        """Register task to database and return database id of the task."""
        raise NotImplementedError

    async def record_run_start(self) -> None:
        """Record in the DB that the crontask has started."""
        raise NotImplementedError

    async def record_run_end(self, response: CronResponse) -> None:
        """Record in the DB that the crontask has ended."""
        raise NotImplementedError

    async def record_stats(self, stats: list[CronStat]) -> None:
        """Record un the DB stats about data to process by this cron."""
        raise NotImplementedError


class TestStorage(CronStorage):
    """Dummy storage used when running test cases."""

    async def record_task(self) -> None:
        """Register task to database and return database id of the task."""

    async def record_run_start(self) -> None:
        """Record in the DB that the crontask has started."""

    async def record_run_end(self, response: CronResponse) -> None:
        """Record in the DB that the crontask has ended."""

    async def record_stats(self, stats: list[CronStat]) -> None:
        """Record un the DB stats about data to process by this cron."""


# StorageT = TypeVar("StorageT", bound=CronStorage)


class CronTask(BaseCronTask):
    """Define a cron task and its storage."""

    storage_class: ClassVar[type[CronStorage]] = CronStorage
    storage: CronStorage

    def __init__(self, **kwargs: Any) -> None:
        """Initialize cron task and storage."""
        super().__init__(**kwargs)
        self.storage = self.storage_class(task=self)

    def get_queryset(self, *, batch_size: Optional[int] = None, **kwargs: Any) -> Any:
        """Fetch the list of elements to process."""
        raise NotImplementedError

    async def process(self, *, batch_size: int = 100, **kwargs: Any) -> Any:
        """Run the cron task and process elements."""
        raise NotImplementedError

    async def get_stats(self) -> list[CronStat]:
        """Give stats about the number of elements left to process."""
        raise NotImplementedError

    async def handle_process(self, *args: Any, **kwargs: Any) -> CronResponse:
        """Run the task and save meta info to Airtable."""

        # Record task
        await self.storage.record_task()

        # Record run start
        await self.storage.record_run_start()

        # Run the task
        response = await super().handle_process(*args, **kwargs)

        # Record run end
        await self.storage.record_run_end(response=response)

        # Compute stats
        stats = await self.get_stats()

        # Record stats
        await self.storage.record_stats(stats=stats)

        return response

    async def test_process(self, filter_queryset: Callable[[FindMany[Any]], FindMany[Any]]) -> CronResponse:
        """Call this method to launch the task in test cases."""
        self.storage = TestStorage(task=self)
        return await super().test_process(filter_queryset)


def register_crontask(
    crontask_class: type[CronTask],
    schedule: celery.schedules.crontab,
    kwargs: Optional[dict[str, Any]] = None,
) -> CronTask:
    """Register a task on the worker servers."""
    return crontask_class(schedule=schedule, kwargs=kwargs or {})


class HealthCheckCron(CronTask):
    """Send a heartbeat signal to better stack to notify that the cron has finished processing."""

    storage_class: ClassVar[type[CronStorage]] = TestStorage

    def get_queryset(self, *, batch_size: Optional[int] = None, **kwargs: Any) -> list[int]:
        """Fetch the list of elements to process."""
        return []

    async def process(self, *, batch_size: int = 100, **kwargs: Any) -> dict[str, int]:
        """Perform heartbeat signal."""
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.head(kwargs["heartbeat_url"])
        assert response.status_code == rest_status.HTTP_200_OK
        return {"status": response.status_code}

    async def get_stats(self) -> list[CronStat]:
        """Give stats about the number of elements left to process."""
        return []



# pylint: disable=too-many-arguments, protected-access, too-many-locals

"""
Utils.

Re-usable methods and functions for all test cases.
"""
from __future__ import annotations

import random
import string
from typing import Any

import httpx
from rich import print  # pylint: disable=redefined-builtin

from sap.rest.rest_exceptions import RestErrorMap


def generate_random_string(length: int) -> str:
    """Generate a random string of a given length."""
    return "".join(random.choices(string.ascii_lowercase + string.digits, k=length))


def generate_random_email(domain: str = "yopmail.net", length: int = 9) -> str:
    """Get a random email to be using in test cases."""
    random_string = "".join(random.choice(string.ascii_lowercase) for _ in range(length))
    return "trellis-test." + random_string + "@" + domain


class PytestState(dict[str, Any]):
    """Store a global state for the current batch of tests."""


CACHED_DATA: dict[str, dict[str, Any]] = {}


# def stringify_request_key(obj: Mapping[str, Any] | Mapping[bytes, Any] | Sequence[Any] | httpx.Headers | httpx.QueryParams | None) -> str:
def stringify_request_key(obj: Any) -> str:
    """Return a human readable string transformation of dict inspired by pytest parametrization."""
    if obj is None:
        return "None"
    items: list[str] = []
    for k, v in obj.items():
        if isinstance(v, dict):
            v_str = stringify_request_key(v)
        else:
            v_str = str(v)
        items.append(f"{k}_{v_str}")

    return "-".join(items)


original__https__request = httpx.AsyncClient.request


async def cache__httpx__request(
    self: httpx.AsyncClient,
    method: str,
    url: httpx._types.URLTypes,
    *,
    content: httpx._types.RequestContent | None = None,
    data: httpx._types.RequestData | None = None,
    files: httpx._types.RequestFiles | None = None,
    json: Any | None = None,
    params: httpx._types.QueryParamTypes | None = None,
    headers: httpx._types.HeaderTypes | None = None,
    cookies: httpx._types.CookieTypes | None = None,
    auth: httpx._types.AuthTypes | httpx._client.UseClientDefault | None = httpx._client.USE_CLIENT_DEFAULT,
    follow_redirects: bool | httpx._client.UseClientDefault = httpx._client.USE_CLIENT_DEFAULT,
    timeout: httpx._types.TimeoutTypes | httpx._client.UseClientDefault = httpx._client.USE_CLIENT_DEFAULT,
    extensions: httpx._types.RequestExtensions | None = None,
) -> httpx.Response:
    """Return cached request if exists otherwise perform HTTP request."""
    request = self.build_request(
        method=method,
        url=url,
        content=content,
        data=data,
        files=files,
        json=json,
        params=params,
        headers=headers,
        cookies=cookies,
        timeout=timeout,
        extensions=extensions,
    )

    request_key = ":".join(
        [
            method,
            str(url),
            str(content),
            stringify_request_key(data),
            str(files),
            stringify_request_key(json),
            stringify_request_key(params),
            stringify_request_key(headers),
        ]
    )

    if request_key in CACHED_DATA:
        result_cached: dict[str, Any] = CACHED_DATA[request_key]
        response = httpx.Response(**result_cached, request=request)
        if result_cached["status_code"] >= 400:
            raise RestErrorMap[result_cached["status_code"]](
                response=response, request=request, data=result_cached["json"]
            )
        return response

    print(f"NON-CACHED HTTPX REQUEST {request_key=} result=>")
    # breakpoint()
    response = await self.send(request, auth=auth, follow_redirects=follow_redirects)

    response_data: dict[str, Any] = {
        "status_code": response.status_code,
        # "headers": response.headers,
    }
    if response.content:
        if "application/json" in response.headers.get("content-type", ""):
            response_data["json"] = response.json()
        else:
            response_data["content"] = response.content
        print(response_data)
    return response



# pylint: disable=useless-import-alias

"""
Typing.

Provide helper for common typing syntax.
"""

try:
    from types import UnionType as UnionType
except ImportError:
    from typing import Union as UnionType  # type: ignore



# pylint: disable=too-many-arguments, too-many-positional-arguments

"""REST testing library utils."""

from __future__ import annotations

import typing
from base64 import b64encode

from async_asgi_testclient import TestClient
from requests.models import Response

from fastapi import status

from AppMain.asgi import app
from sap.beanie.document import DocT
from sap.fastapi.pagination import PaginatedResponse
from sap.fastapi.user import UserMixin

if typing.TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict


class Headers(typing.TypedDict, total=False):
    """Format Request headers."""

    Authorization: str


async def get_headers(user: UserMixin) -> Headers:
    """Get default headers for api authentication."""
    auth_key = await user.get_auth_key()
    basic_auth = b64encode((f"{auth_key}:").encode("utf-8")).decode("ascii")
    headers: Headers = {"Authorization": f"Basic {basic_auth}"}
    return headers


async def assert_response_forbidden(response: Response) -> bool:
    """Check that a request was rejected to due to lack sufficient authorization."""
    assert response.status_code in [
        status.HTTP_401_UNAUTHORIZED,
        status.HTTP_403_FORBIDDEN,
        # status.HTTP_404_NOT_FOUND,
        status.HTTP_405_METHOD_NOT_ALLOWED,
    ], response.content

    return True


async def assert_rest_can_list(
    base_url: str,
    sample: typing.Optional[dict[str, typing.Any]] = None,
    user: typing.Optional[UserMixin] = None,
    roles: typing.Optional[list[str]] = None,
) -> bool:
    """Retrieve an item through its ID."""
    headers = await get_headers(user) if user else {}

    async with TestClient(app, headers=headers) as client:
        response = await client.get(base_url)

    async def assert_response_authorized() -> bool:
        """Verify that the action was successfully performed."""

        assert response.status_code == status.HTTP_200_OK, response.content
        assert sample

        # Ensure that the response data has the right format
        # Ensure that the output data matches the input
        response_data: PaginatedResponse = response.json()
        assert sample.keys() == response_data["data"][0].keys()

        return True

    if roles is None or (user and user.role in roles):
        # roles is None => Can perform action without authentication
        # user.role in roles => Authenticated user can perform action if they have permission
        return await assert_response_authorized()

    return await assert_response_forbidden(response)


async def assert_rest_can_retrieve(
    base_url: str,
    item_id: str,
    sample: typing.Optional[dict[str, typing.Any]] = None,
    user: typing.Optional[UserMixin] = None,
    roles: typing.Optional[list[str]] = None,
) -> bool:
    """Retrieve an item through its ID."""
    headers = await get_headers(user) if user else {}

    async with TestClient(app, headers=headers) as client:
        response = await client.get(f"{base_url}{item_id}/")

    async def assert_response_authorized() -> bool:
        """Verify that the action was successfully performed."""
        assert response.status_code == status.HTTP_200_OK, response.content
        assert sample

        # Ensure that the response data has the right format
        response_data: dict[str, typing.Any] = response.json()
        assert (
            sample.keys() == response_data.keys()
        ), f"WrongObject expected={sample.keys()} received={response_data.keys()}"

        # from rich import print
        # print(response_data)

        for key, value in sample.items():
            if value is None:
                continue
            assert type(value) == type(  # pylint: disable=unidiomatic-typecheck
                response_data[key]
            ), f"TypeMismatch key={key} expected={type(value)} received={type(response_data[key])}"
            if isinstance(value, dict):
                assert (
                    value.keys() == response_data[key].keys()
                ), f"WrongEmbeddedObject key={key} expected={value.keys()} received={response_data[key].keys()}"

        return True

    if roles is None or (user and user.role in roles):
        # roles is None => Can perform action without authentication
        # user.role in roles => Authenticated user can perform action
        return await assert_response_authorized()

    return await assert_response_forbidden(response)


async def assert_rest_can_create(
    base_url: str,
    sample: dict[str, typing.Any],
    variant_good: typing.Optional[dict[str, typing.Any]] = None,
    variant_bad: typing.Optional[dict[str, typing.Any]] = None,
    user: typing.Optional[UserMixin] = None,
    roles: typing.Optional[list[str]] = None,
) -> bool:
    """Update an item using its ID."""
    headers = await get_headers(user) if user else {}

    variant_good = variant_good or {}
    variant_bad = variant_bad or {}

    async with TestClient(app, headers=headers) as client:
        response_good = await client.post(base_url, json=sample | variant_good)
        response_bad = await client.post(base_url, json=sample | variant_bad)

    async def assert_response_authorized() -> bool:
        """Verify that the action was successfully performed."""
        assert variant_good
        assert variant_bad

        assert response_good.status_code == status.HTTP_201_CREATED, response_good.content
        assert response_bad.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY, response_bad.content

        # Ensure that the response data has the right format
        response_good_data: dict[str, typing.Any] = response_good.json()
        for key, value in variant_good.items():
            # some case are only accepted at creation for example admin_email when creating agency
            if key in response_good_data:
                error_message = f"Updating data failed {key=} expected={value} found={response_good_data[key]}"
                assert response_good_data[key] == value, error_message

        response_bad_data: dict[str, typing.Any] = response_bad.json()
        details: typing.Union[list["ErrorDict"], str] = response_bad_data["detail"]

        if isinstance(details, list):
            for key, value in variant_bad.items():
                error_message = (
                    f"Updating data was not rejected as expected {key=} {value=} receive={response_bad_data}"
                )
                assert any(key in detail["loc"] for detail in details), error_message

        return True

    if roles is None or (user and user.role in roles):
        # roles is None => Can perform action without authentication
        # user.role in roles => Authenticated user can perform action
        return await assert_response_authorized()

    return await assert_response_forbidden(response_good) and await assert_response_forbidden(response_bad)


async def assert_rest_can_update(
    base_url: str,
    item_id: str,
    variant_good: typing.Optional[dict[str, typing.Any]] = None,
    variant_bad: typing.Optional[dict[str, typing.Any]] = None,
    user: typing.Optional[UserMixin] = None,
    roles: typing.Optional[list[str]] = None,
) -> bool:
    """Update an item using its ID."""
    headers = await get_headers(user) if user else {}

    variant_good = variant_good or {}
    variant_bad = variant_bad or {}

    async with TestClient(app, headers=headers) as client:
        data_initial = (await client.get(f"{base_url}{item_id}/")).json()
        response_good = await client.put(f"{base_url}{item_id}/", json=data_initial | variant_good)
        response_bad = await client.put(f"{base_url}{item_id}/", json=data_initial | variant_bad)

    async def assert_response_authorized() -> bool:
        """Verify that the action was successfully performed."""
        assert variant_good
        assert variant_bad

        assert response_good.status_code == status.HTTP_202_ACCEPTED, response_good.content
        assert response_bad.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY, response_bad.content

        # Ensure that the response data has the right format
        response_good_data: dict[str, typing.Any] = response_good.json()
        for key, value in variant_good.items():
            error_message = f"Updating data failed {key=} expected={value} found={response_good_data[key]}"
            assert response_good_data[key] == value, error_message

        response_bad_data: dict[str, typing.Any] = response_bad.json()
        for key, value in variant_bad.items():
            details: list["ErrorDict"] = response_bad_data["detail"]
            error_message = f"Updating data was not rejected as expected {key=} {value=} receive={response_bad_data}"
            assert any(key in detail["loc"] for detail in details), error_message

        return True

    if roles is None or (user and user.role in roles):
        # roles is None => Can perform action without authentication
        # user.role in roles => Authenticated user can perform action
        return await assert_response_authorized()

    return await assert_response_forbidden(response_good) and await assert_response_forbidden(response_bad)


async def assert_rest_can_destroy(
    base_url: str,
    item_id: str,
    item: DocT,
    user: typing.Optional[UserMixin] = None,
    roles: typing.Optional[list[str]] = None,
) -> bool:
    """Delete an item."""
    headers = await get_headers(user) if user else {}

    async with TestClient(app, headers=headers) as client:
        response = await client.delete(f"{base_url}{item_id}/")

    async def assert_response_authorized() -> bool:
        """Verify that the action was successfully performed."""
        assert response.status_code == status.HTTP_204_NO_CONTENT
        assert not await item.get(item_id)
        return True

    if roles is None or (user and user.role in roles):
        # roles is None => Can perform action without authentication
        # user.role in roles => Authenticated user can perform action
        return await assert_response_authorized()

    return await assert_response_forbidden(response)



"""
Exceptions.

Group common logic exceptions such as ValidationError
"""

from fastapi import HTTPException, status

from sap.beanie.exceptions import Object404Error


class HTTPError(HTTPException):
    """Subclassing HTTPException."""

    status_code: int
    detail: str

    def __init__(self, detail: str = ""):
        """Init exception."""
        super().__init__(status_code=self.status_code, detail=detail or self.detail)


class Validation422Error(HTTPError):
    """Raise when parameter validation fails."""

    status_code: int = status.HTTP_422_UNPROCESSABLE_ENTITY
    detail: str = "The data submitted is invalid."


class Validation400Error(HTTPError):
    """Raise when data validation fails."""

    status_code: int = status.HTTP_400_BAD_REQUEST
    detail: str = "The data submitted has been rejected."


__all__ = [
    "Object404Error",
    "Validation400Error",
    "Validation422Error",
]



"""
Auth.

Utils to handle user authentication and permissions.
"""

from __future__ import annotations

import base64
import binascii
import time
from typing import ClassVar, TypeVar, Union

import jwt
import typing_extensions

from fastapi import Request, Response
from fastapi.exceptions import HTTPException
from pydantic import BaseModel

# from starlette.authentication import AuthCredentials, AuthenticationBackend, AuthenticationError, BaseUser
# from starlette.requests import HTTPConnection
from starlette.status import HTTP_307_TEMPORARY_REDIRECT as HTTP_307
from starlette.status import HTTP_401_UNAUTHORIZED as HTTP_401

from AppMain.settings import AppSettings
from sap.beanie.document import Document
from sap.beanie.exceptions import Object404Error

UserT = TypeVar("UserT", bound=Document)
UserViewT: typing_extensions.TypeAlias = Union[Document, BaseModel, object]


class JWTAuth:
    """JWT cookie authentication utils.

    This can be used to login/logout user with persistent sessions.
    Mainly useful for Web Apps. For API, it is better to use Basic Authentication in Headers.
    """

    auth_login_url: ClassVar[str] = "/pages/auth/login/"
    auth_cookie_key: ClassVar[str] = "user_session"
    auth_cookie_expires: ClassVar[int] = 60 * 60 * 12  # expiration = 12 hours
    user_model: type[Document]

    def __init__(self, user_model: type[UserT]) -> None:
        """Initialize the JWT auth helper.

        :param user_model: The User model class.
        """
        super().__init__()
        self.user_model = user_model

    def get_auth_login_url(self, request: Request) -> str:
        """Retrieve the login url where user are redirect in case of auth failure."""
        return self.auth_login_url

    def get_auth_cookie_key(self, request: Request) -> str:
        """Retrieve key used to define the authentication cookie."""
        return self.auth_cookie_key

    def get_auth_cookie_expires(self) -> int:
        """Retrieve validity in seconds of the authentication cookie."""
        return self.auth_cookie_expires

    def create_token(self, user: UserT) -> str:
        """Get JWT temporary token."""
        expires = self.get_auth_cookie_expires()
        jwt_data = {"exp": int(time.time()) + expires, "user_id": str(user.id)}
        return jwt.encode(payload=jwt_data, key=AppSettings.CRYPTO_SECRET, algorithm="HS256")

    async def find_user(self, jwt_token: str) -> Document:
        """
        Verify that JWT token is valid.

        :param jwt_token: The lifespan of the token in seconds.
        """

        # Raises: jwt.exceptions.InvalidTokenError => Token has expired or is invalid
        jwt_data = jwt.decode(jwt_token, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
        if "user_id" not in jwt_data:
            raise jwt.exceptions.InvalidAudienceError("Cannot read user_id.")

        # Raises: Object404Error => User cannot be found
        return await self.user_model.get_or_404(jwt_data["user_id"])

    async def login(self, response: Response, request: Request, user: UserT) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.set_cookie(
            key=self.get_auth_cookie_key(request),
            value=self.create_token(user=user),
            httponly=True,
        )
        return response

    async def logout(self, response: Response, request: Request) -> Response:
        """Create a persistent cookie based session for the authenticated user."""
        response.delete_cookie(key=self.get_auth_cookie_key(request), httponly=True)
        return response

    async def authenticate(self, request: Request) -> UserT:
        """Provide the authenticated user to views that require it."""
        try:
            jwt_token = request.cookies[self.get_auth_cookie_key(request)]
        except KeyError as exc:
            raise HTTPException(HTTP_307, headers={"Location": self.get_auth_login_url(request)}) from exc

        try:
            return await self.find_user(jwt_token=jwt_token)
        except (Object404Error, jwt.exceptions.InvalidTokenError) as exc:
            raise HTTPException(HTTP_307, headers={"Location": self.get_auth_login_url(request)}) from exc


# class JWTAuthBackend(AuthenticationBackend, JWTAuth):
#     """Starlette Backend to authenticate use through JWT Token in Cookies."""

#     async def authenticate(self, request: HTTPConnection) -> UserT:
#         """Authenticate the user using Cookies."""
#         cookie_key: str = self.get_auth_cookie_key(request=request)
#         if cookie_key not in request.cookies:
#             raise AuthenticationError("Unable to find JWT cookie")

#         jwt_cookie = request.cookies[cookie_key]

#         try:
#             jwt_data = jwt.decode(jwt_cookie, key=AppSettings.CRYPTO_SECRET, algorithms=["HS256"])
#         except jwt.exceptions.InvalidTokenError as exc:
#             raise AuthenticationError("Invalid JWT token") from exc

#         user = await self.user_model.get_or_404(jwt_data["user_id"])

#         return user
#         # return AuthCredentials([user.get_scopes()]), user


class BasicAuth:
    """Basic authentication utils.

    This can be used to authenticate user with through `Authorization` header using the Basic protocol.
    Mainly useful for API access. For web app, check JWTAuth which support persistent cookie session.
    """

    user_model: type[UserViewT] | None
    auth_key_attribute: ClassVar[str] = "auth_key"

    def __init__(self, user_model: type[UserViewT] | None = None) -> None:
        """Initialize the auth helper.

        :param user_model: The User model class.
        """
        super().__init__()
        self.user_model = user_model

    def get_auth_key_attribute(self) -> str | None:
        """Retrieve name of the `auth_key` attribute use to verify user."""
        return self.auth_key_attribute

    async def retrieve_user(self, user_key: str) -> UserViewT:
        """Retrieve a user using authorization key."""
        if self.user_model and issubclass(self.user_model, Document):
            if auth_key_name := self.get_auth_key_attribute():
                auth_key = getattr(self.user_model, auth_key_name)
                return await self.user_model.find_one_or_404(auth_key == user_key)

            return await self.user_model.get_or_404(user_key)
        raise NotImplementedError

    async def authenticate(self, request: Request) -> UserViewT:
        """Provide the authenticated user to views that require it."""

        header_auth: str | None = request.headers.get("Authorization")
        if not header_auth:
            raise HTTPException(HTTP_401, detail="Authentication required")

        scheme, credentials = header_auth.split()
        if scheme.lower() != "basic":
            raise HTTPException(HTTP_401, detail="Only basic authorization is supported")

        try:
            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
            raise HTTPException(HTTP_401, detail="Error while decoding basic auth credentials") from exc

        username, _, pwd = decoded.partition(":")
        user_key = username or pwd

        try:
            return await self.retrieve_user(user_key)
        except (Object404Error, jwt.exceptions.InvalidTokenError) as exc:
            raise HTTPException(HTTP_401, detail="Invalid basic auth credentials") from exc



"""
Pydantic.

This package regroups all common helpers to work with Pydantic
"""

from datetime import datetime, timezone
from typing import TypeVar

import pydantic


def datetime_utcnow() -> datetime:
    """Replace deprecated datetime.utcnow."""
    return datetime.now(timezone.utc)


ModelT = TypeVar("ModelT", bound=pydantic.BaseModel)



"""
Stat Serializer.

A serializer that helps normalize the format of statistics objects
"""

import typing
from decimal import Decimal, DecimalException

import pydantic


class StatSerializer(pydantic.BaseModel):
    """A stat object used to display a stat card."""

    name: str
    description: str
    value: int
    total: typing.Optional[int] = None
    percent: typing.Optional[Decimal] = None
    view: typing.Optional[str] = None  # URL the to the page to view the data with filter query

    @pydantic.model_validator(mode="before")
    @classmethod
    def set_percent(cls, values: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """Calculate the percentage."""
        if values.get("total") is None:
            return values

        try:
            percent = Decimal(values["value"]) * 100 / values["total"]
        except (ZeroDivisionError, DecimalException):
            percent = 0

        values["percent"] = int(percent)
        return values



"""
Exceptions.

Group exceptions specific to database error such as ObjectNotFound or IntegrityError
"""

from fastapi.exceptions import HTTPException


class Object404Error(HTTPException):
    """Raise when querying DB returns empty result."""

    status_code: int = 404
    detail: str = "Object not found"

    def __init__(self, detail: str = ""):
        """Init exception."""
        super().__init__(status_code=404, detail=detail or self.detail)
